#	fkt_header.S
#
#
#	Copyright (C) 2000, 2001  Robert D. Russell -- rdr@unh.edu
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
###	08-Nov-00	rdr		change SMP handling to avoid spin locks
###	11-Jul-00	rdr		change __SMP__ to CONFIG_SMP for 2.4.0
#
#	fkt = Fast Kernel Tracing
#
#	uses 3 externals:
#		unsigned fkt_active;			0 when inactive, non-zero when active
#		unsigned *fkt_next_slot;		pointer to next free slot in buffer
#		unsigned *fkt_limit_slot;		pointer to slot just beyond end of buffr
#
#
#	call is:	fkt_header( header, p1, p2, ... );
#	where:
#				header low byte has total number of bytes in this probe entry
#							= (3 + number_of_parameters)*4
#				header high 3 bytes have entry identification code
#
#
#	format of one probe record in buffer is:
#				int 1		low 32-bits of cycle clock
#				int 2		low half gets 16-bit pid of current process
#							high half gets 16-bit cpu number
#				int 3		full 32-bit header from caller
#				int 4		full 32-bit p1 from caller
#				int 5		full 32-bit p2 from caller
#				...		...
#				int 3+n		full 32-bit pn from caller
#
#
#	assumptions (and invariants):
#		value in fkt_next_slot is a multiple of 4
#		value in fkt_limit_slot is a multiple of 4
#		value in low byte of header is multiple of 4
#
#
#
#	fkt_active contains the currently active set of probe keys.
#	bit x in fkt_active is 1 if probes in keyset x are active.
#	fkt_active is set  when tracing is actived by fkt_setup();  it is reset
#	to 0 when tracing is deactived by fkt_endup().
#	fkt_active is also reset to 0 by the first thread that fills the buffer.
#

#include <linux/linkage.h>

#	Offsets into struct task_struct -- must agree with <linux/sched.h>
#	CAREFUL! these tend to change from version to version
#	since they are NOT wired down!
processor	= 52			# was 40 in 2.2.12, was 56 in 2.4.0-test2
pid			= 108			# was 92 in 2.2.12, was 108 in 2.4.0-test2


#	Offsets into caller stack when in these routines
caller_code	= 8				# when limit not pushed on TOS
caller_offset = 12			# after limit is pushed on TOS


#	Offsets into probe record
probe_cycles	= 0
probe_pid		= 4
probe_cpu_no	= 6
probe_code		= 8
probe_param1	= 12



#	borrowed from entry.S
#	points reg at current task control block (TCB)
#define GET_CURRENT(reg) \
	movl	%esp, reg; \
	andl	$-8192, reg;


	.text

################################################################################
#
# normal entry point to take a probe
#	at entry	return address is on stack pointed to by esp
#				eax, ecx, edx are undefined
#				ebx is unused
#	on return	eax, ecx, edx are destroyed
#				ebx is unchanged
#

	.align	4
	.globl	fkt_header

fkt_header:
	pushfl							# save processor flags
	cli								# block local interrupts

#ifdef CONFIG_SMP
	movzbl	caller_code(%esp),%edx	# edx gets no of bytes needed by this record
	movl	fkt_next_slot,%eax		# eax gets ptr to first slot for this record
1:
	leal	(%eax,%edx,1),%ecx		# ecx gets ptr to next slot after record
	cmpl	%ecx,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	lock
	cmpxchgl %ecx,fkt_next_slot		# atomic update ptr to nxt slot after record
	jnz		1b						# jump back if we did not get it
	pushl	%ecx					# TOS gets ptr to next slot after record
	movl	%eax,%ecx				# ecx gets ptr to first slot for this record
#else
	movl	fkt_next_slot,%ecx		# ecx gets ptr to first slot for this record
	movzbl	caller_code(%esp),%eax	# eax gets no of bytes needed by this record
	addl	%ecx,%eax				# eax gets ptr to next slot after record
	cmpl	%eax,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	movl	%eax,fkt_next_slot		# update ptr to next slot after record
									# ecx gets ptr to first slot for this record
#endif /* CONFIG_SMP */

	rdtsc							# get 64-bit cpu cycle counter in edx-eax
									# ignore high-order 32 bits in edx
	movl	%eax,probe_cycles(%ecx)	# low-order 32 bits from eax to 1st slot
	GET_CURRENT(%eax)				# point eax at current task control block

#ifdef CONFIG_SMP
	movl	pid(%eax),%edx			# edx gets pid of current process
	movw	%dx,probe_pid(%ecx)		# 2nd slot (lower half) gets pid
	movl	processor(%eax),%eax	# eax get current processor no
	movw	%ax,probe_cpu_no(%ecx)	# 2nd slot (upper half) gets cpu no
	leal	caller_offset(%esp),%edx # edx points at header value in call
#else
	movl	pid(%eax),%eax			# eax gets pid of current process
	movl	%eax,probe_pid(%ecx)	# whole 2nd slot gets pid (top half 0)
	leal	caller_code(%esp),%edx	# edx points at header value in call
#endif /* CONFIG_SMP */

	leal	probe_code(%ecx),%ecx	# ecx points at header slot in record
9:
	movl	(%edx),%eax				# pick up next value from caller
	movl	%eax,(%ecx)				# save that value into next record slot
	leal	4(%edx),%edx			# bump edx to point at next caller value
	leal	4(%ecx),%ecx			# bump ecx to point at next record slot

#ifdef CONFIG_SMP
	cmpl	%ecx,(%esp)				# see if record filled yet
	jne		9b						# if not, jump back to next loop iteration
	popl	%eax					# pop limit off TOS
#else
	cmpl	%ecx,fkt_next_slot		# see if record filled yet
	jne		9b						# if not, jump back to next loop iteration
#endif /* CONFIG_SMP */

6:
	popfl							# restore processor flags, local interrupts
	ret								# return to caller

	.align	4
5:
	movl	$0,fkt_active			# turn all active bits off
	jmp		6b						# leave


################################################################################
#
#	fastest entry point to take a probe for sys calls and traps (with no params)
#	probe record is 12 bytes
#	at entry	return address is on stack pointed to by esp
#				eax contains the code
#				ebx contains a pointer to the TCB
#				ecx is undefined
#				edx is undefined
#	on return	eax, ecx, edx are destroyed
#				ebx is unchanged
#

	.align	4
	.globl	fkt_eax_ebx0

fkt_eax_ebx0:
	pushfl							# save processor flags
	cli								# block local interrupts

#ifdef CONFIG_SMP
	movl	%eax,%edx				# edx gets code (in wrong register at entry)
	movl	fkt_next_slot,%eax		# eax gets ptr to first slot for this record
1:
	leal	12(%eax),%ecx			# ecx gets ptr to next slot after record
	cmpl	%ecx,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	lock
	cmpxchgl %ecx,fkt_next_slot		# atomic update ptr to nxt slot after record
	jnz		1b						# jump back if we did not get it
	movl	%eax,%ecx				# ecx gets ptr to first slot for this record
	movl	%edx,probe_code(%ecx)	# save code
#else
	movl	fkt_next_slot,%ecx		# ecx gets ptr to first slot for this record
	leal	12(%ecx),%edx			# edx gets ptr to next slot after record
	cmpl	%edx,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	movl	%edx,fkt_next_slot		# update ptr to next slot after record
	movl	%eax,probe_code(%ecx)	# save code
#endif /* CONFIG_SMP */

	rdtsc							# get 64-bit cpu cycle counter in edx-eax
									# ignore high-order 32 bits in edx
	movl	%eax,probe_cycles(%ecx)	# low-order 32 bits from eax to 1st slot
	movl	pid(%ebx),%eax			# eax gets pid of current process

#ifdef CONFIG_SMP
	movw	%ax,probe_pid(%ecx)		# 2nd slot (lower half) gets pid
	movl	processor(%ebx),%eax	# eax gets current processor no
	movw	%ax,probe_cpu_no(%ecx)	# 2nd slot (upper half) gets cpu no
#else
	movl	%eax,probe_pid(%ecx)	# whole 2nd slot gets pid (top half 0)
#endif /* CONFIG_SMP */

6:
	popfl							# restore processor flags, local interrupts
	ret								# return to caller

	.align	4
5:
	movl	$0,fkt_active			# turn all active bits off
	jmp		6b						# leave


################################################################################
#
#	faster entry point to take a probe for IRQs and other probes with no params
#	probe record is 12 bytes
#	at entry	return address is on stack pointed to by esp
#				eax contains the code
#				ebx is unused
#				ecx is undefined
#				edx is undefined
#	on return	eax, ecx, edx are destroyed
#				ebx is unchanged
#

	.align	4
	.globl	fkt_eax0

fkt_eax0:
	pushfl							# save processor flags
	cli								# block local interrupts

#ifdef CONFIG_SMP
	movl	%eax,%edx				# edx gets code (in wrong register at entry)
	movl	fkt_next_slot,%eax		# eax gets ptr to first slot for this record
1:
	leal	12(%eax),%ecx			# ecx gets ptr to next slot after record
	cmpl	%ecx,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	lock
	cmpxchgl %ecx,fkt_next_slot		# atomic update ptr to nxt slot after record
	jnz		1b						# jump back if we did not get it
	movl	%eax,%ecx				# ecx gets ptr to first slot for this record
	movl	%edx,probe_code(%ecx)	# save code
#else
	movl	fkt_next_slot,%ecx		# ecx gets ptr to first slot for this record
	leal	12(%ecx),%edx			# edx gets ptr to next slot after record
	cmpl	%edx,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	movl	%edx,fkt_next_slot		# update ptr to next slot after record
	movl	%eax,probe_code(%ecx)	# save code
#endif /* CONFIG_SMP */

	rdtsc							# get 64-bit cpu cycle counter in edx-eax
									# ignore high-order 32 bits in edx
	movl	%eax,probe_cycles(%ecx)	# low-order 32 bits from eax to 1st slot
	GET_CURRENT(%eax)				# point eax at current task control block

#ifdef CONFIG_SMP
	movl	pid(%eax),%edx			# edx gets pid of current process
	movw	%dx,probe_pid(%ecx)		# 2nd slot (lower half) gets pid
	movl	processor(%eax),%eax	# eax gets current processor no
	movw	%ax,probe_cpu_no(%ecx)	# 2nd slot (upper half) gets cpu no
#else
	movl	pid(%eax),%eax			# eax gets pid of current process
	movl	%eax,probe_pid(%ecx)	# whole 2nd slot gets pid (top half 0)
#endif /* CONFIG_SMP */

6:
	popfl							# restore processor flags, local interrupts
	ret								# return to caller

	.align	4
5:
	movl	$0,fkt_active			# turn all active bits off
	jmp		6b						# leave


################################################################################
#
#	faster entry point to take a probe with one parameter
#	expects 1 parameter in %edx, and code in %eax
#	probe record is 16 bytes
#	at entry	return address is on stack pointed to by esp
#				eax contains the code
#				ebx is unused
#				ecx is undefined
#				edx contains the parameter
#	on return	eax, ecx, edx are destroyed
#				ebx is unchanged
#

	.align	4
	.globl	fkt_eax1

fkt_eax1:
	pushfl							# save processor flags
	cli								# block local interrupts
	pushl	%eax					# save the code because we need eax

#ifdef CONFIG_SMP
	movl	fkt_next_slot,%eax		# eax gets ptr to first slot for this record
1:
	leal	16(%eax),%ecx			# ecx gets ptr to next slot after record
	cmpl	%ecx,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	lock
	cmpxchgl %ecx,fkt_next_slot		# atomic update ptr to nxt slot after record
	jnz		1b						# jump back if we did not get it
	movl	%eax,%ecx				# ecx gets ptr to first slot for this record
#else
	movl	fkt_next_slot,%ecx		# ecx gets ptr to first slot for this record
	leal	16(%ecx),%eax			# eax gets ptr to next slot after record
	cmpl	%eax,fkt_limit_slot		# are we off the end of the buffer?
	jb		5f						# jump if below (not enough free bytes)
	movl	%eax,fkt_next_slot		# update ptr to next slot after record
#endif /* CONFIG_SMP */

	popl	%eax					# get back code
	movl	%eax,probe_code(%ecx)	# save code
	movl	%edx,probe_param1(%ecx)	# save first (only) parameter
	rdtsc							# get 64-bit cpu cycle counter in edx-eax
									# ignore high-order 32 bits in edx
	movl	%eax,probe_cycles(%ecx)	# low-order 32 bits from eax to 1st slot
	GET_CURRENT(%eax)				# point eax at current task control block

#ifdef CONFIG_SMP
	movl	pid(%eax),%edx			# edx gets pid of current process
	movw	%dx,probe_pid(%ecx)		# 2nd slot (lower half) gets pid
	movl	processor(%eax),%eax	# eax gets current processor no
	movw	%ax,probe_cpu_no(%ecx)	# 2nd slot (upper half) gets cpu no
#else
	movl	pid(%eax),%eax			# eax gets pid of current process
	movl	%eax,probe_pid(%ecx)	# whole 2nd slot gets pid (top half 0)
#endif /* CONFIG_SMP */

6:
	popfl							# restore processor flags, local interrupts
	ret								# return to caller

	.align	4
5:
	movl	$0,fkt_active			# turn all active bits off
	popl	%eax					# get back code
	jmp		6b						# leave


################################################################################
#
	.globl	fkt_trailer				# to see size of code generated
fkt_trailer:
	.size   fkt_header,fkt_trailer-fkt_header
